unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ComCtrls, StdCtrls, Clipbrd, Palettes, Buttons;

const PaletteFileBaseName='colsel.pal';

type TSlope=array[0..359]of integer;
     TConversionDirection=(RGB2FHS,FHS2RGB);
     TColorComponent=(ccR,ccG,ccB);
     str2=string[2];
     str7=string[7];

type
  TMainForm = class(TForm)
    ColSelPanel: TPanel;
    ColPick: TImage;
    Preview: TPanel;
    PalSelect: TComboBox;
    PalBox: TListBox;
    ColFHS: TGroupBox;
    Label1: TLabel;
    ctF: TTrackBar;
    cnF: TEdit;
    cdF: TUpDown;
    ctH: TTrackBar;
    Label2: TLabel;
    cnH: TEdit;
    cdH: TUpDown;
    ctS: TTrackBar;
    Label3: TLabel;
    cnS: TEdit;
    cdS: TUpDown;
    ColRGB: TGroupBox;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    ctR: TTrackBar;
    cnR: TEdit;
    cdR: TUpDown;
    ctG: TTrackBar;
    cnG: TEdit;
    cdG: TUpDown;
    ctB: TTrackBar;
    cnB: TEdit;
    cdB: TUpDown;
    csHTML: TEdit;
    cpHTML: TButton;
    Label7: TLabel;
    Label8: TLabel;
    csDelphi: TEdit;
    cpDelphi: TButton;
    Bevel1: TBevel;
    BAdd: TBitBtn;
    BDel: TBitBtn;
    ColorName: TEdit;
    BScreen: TButton;
    procedure varint(Sender: TObject);
    procedure ColPickMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ColPickMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ColPickMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure NumbersOnly(Sender: TObject; var Key: Char);
    procedure cnFChange(Sender: TObject);
    procedure cnHSChange(Sender: TObject);
    procedure ctChange(Sender: TObject);
    procedure cnRGBChange(Sender: TObject);
    procedure cpHTMLClick(Sender: TObject);
    procedure cpDelphiClick(Sender: TObject);
    procedure HexOnly(Sender: TObject; var Key: Char);
    procedure PalSelectChange(Sender: TObject);
    procedure PalBoxDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure PalBoxClick(Sender: TObject);
    procedure ColorNameChange(Sender: TObject);
    procedure BAddClick(Sender: TObject);
    procedure BDelClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure BScreenClick(Sender: TObject);
  private
    { Private-Deklarationen}
    PaletteFileName:string;
    F,H,S:integer;
    RV,GV,BV:byte;
    SlopeR,SlopeG,SlopeB:TSlope;
    Circ,Grad:TBitmap;
    HoldMouseOnCircle,HoldMouseOnGradient:boolean;
    CircleMinRad,CircleMaxRad:integer;
    Lock:boolean;
    CurrentPal:integer;
    procedure CreateCircle;
    procedure ChangeHue;
    procedure ChangeColor(Dir:TConversionDirection);
    procedure ConvRGB2FHS;
    procedure GetHexColor(const Raw:str7);
    procedure LoadPalette;
  public
    { Public-Deklarationen}
    procedure ImportColor(Color:TColor);
  end;

var
  MainForm: TMainForm;

const CircleMinR2=09000;
      CircleMaxR2=14000;
      CrosshairSize=8;

implementation

uses ScreenShot;

{$R *.DFM}

const rad2deg=180.0/pi;
      hex:string[16]='0123456789ABCDEF';

procedure CreateBaseSlope(var Slope:TSlope);
var i:integer;w:word;
begin
  for i:=0 to 359 do Slope[i]:=0;
  for i:=0 to 60 do begin
    Slope[i]:=32768;
    Slope[359-i]:=32768;
  end;
  for i:=0 to 60 do begin
    w:=round(i/60.0*32768.0);
    Slope[120-i]:=w;
    Slope[240+i]:=w;
  end;
end;

procedure RotateSlope120(var src,dest:TSlope);
var a,b:integer;
begin
  b:=120;
  for a:=0 to 359 do begin
    dest[b]:=src[a];
    inc(b);if b>359 then dec(b,360);
  end;
end;

procedure TMainForm.varint(Sender: TObject);
var cnt:integer; LastSlash:integer;
begin
//AllocConsole;
  PaletteFileName:=ParamStr(0);
  LastSlash:=length(PaletteFileName);
  while PaletteFileName[LastSlash]<>'\' do dec(LastSlash);
  PaletteFileName:=copy(PaletteFileName,1,LastSlash)+PaletteFileBaseName;

  Lock:=true;
  Grad:=TBitmap.Create;
  with Grad do begin
    PixelFormat:=pf24bit;
    Width:=128;
    Height:=128;
  end;
  Circ:=TBitmap.Create;
  with Circ do begin
    PixelFormat:=pf24bit;
    Width:=240;
    Height:=240;
  end;
  CircleMinRad:=round(sqrt(CircleMinR2)-3.0);
  CircleMaxRad:=round(sqrt(CircleMaxR2)+5.0);
  ctF.Tag:=Integer(Pointer(@cnF));
  ctH.Tag:=Integer(Pointer(@cnH));
  ctS.Tag:=Integer(Pointer(@cnS));
  ctR.Tag:=Integer(Pointer(@cnR));
  ctG.Tag:=Integer(Pointer(@cnG));
  ctB.Tag:=Integer(Pointer(@cnB));

  PalData[0].LoadFromFile(PaletteFileName);
  with PalSelect,Items do begin
    Clear;
    for cnt:=0 to (NumPalettes-1) do
      Add(PalData[cnt].Name);
    ItemIndex:=0;
  end;
  CurrentPal:=0;
  LoadPalette;

  CreateBaseSlope(SlopeR);
  RotateSlope120(SlopeR,SlopeG);
  RotateSlope120(SlopeG,SlopeB);
  CreateCircle;
  F:=0;H:=0;S:=0;
  HoldMouseOnCircle:=false;
  HoldMouseOnGradient:=false;
  Lock:=false;
  ChangeHue;
end;

var RefLine:array[0..383]of byte;
procedure TMainForm.ChangeHue;
var CurrLine:PByteArray;
    CurrPos,RefPos:^byte;
    X,Y:integer;
    X2,Y2:integer;
    R,G,B:integer;
    HadLock:boolean;
begin
  HadLock:=Lock; Lock:=true;
  RefPos:=@RefLine[0];
  for X:=0 to 127 do begin
    X2:=(X SHL 1) OR (X SHR 6);
    R:=SlopeR[F]*X2 SHR 15+255-X2;
    G:=SlopeG[F]*X2 SHR 15+255-X2;
    B:=SlopeB[F]*X2 SHR 15+255-X2;
    RefPos^:=B;inc(RefPos);
    RefPos^:=G;inc(RefPos);
    RefPos^:=R;inc(RefPos);
  end;
  for Y:=0 to 127 do begin
    Y2:=(Y SHL 1) OR (Y SHR 6);
    CurrLine:=Grad.ScanLine[127-Y];
    CurrPos:=@CurrLine[0];
    RefPos:=@RefLine[0];
    for X:=0 to 383 do begin
      CurrPos^:=RefPos^*Y2 DIV 255;
      inc(CurrPos);inc(RefPos);
    end;
  end;
  if not HadLock then ChangeColor(FHS2RGB);
end;

procedure TMainForm.ChangeColor(Dir:TConversionDirection);
var R,G,B,I:integer;
    vX,vY:integer;
    sR,sG,sB:str2;
begin
  Lock:=true;
  if Dir=FHS2RGB then begin
    I:=S SHL 7;
    R:=SlopeR[F]*I SHR 15+32768-I;
    G:=SlopeG[F]*I SHR 15+32768-I;
    B:=SlopeB[F]*I SHR 15+32768-I;
    RV:=Integer(R*H SHR 15);
    GV:=Integer(G*H SHR 15);
    BV:=Integer(B*H SHR 15);
  end else begin
    I:=F;
    ConvRGB2FHS;
    if I<>F then ChangeHue;
  end;
  with ColPick.Canvas do begin
    Draw(0,0,Circ);
    Draw(56,56,Grad);
    Pen.Mode:=pmCopy;
    MoveTo(120-round(CircleMinRad*sin(F/rad2deg)),
           120-round(CircleMinRad*cos(F/rad2deg)));
    LineTo(120-round(CircleMaxRad*sin(F/rad2deg)),
           120-round(CircleMaxRad*cos(F/rad2deg)));
    Pen.Mode:=pmNotXor;
    vX:=056+S SHR 1;
    vY:=183-H SHR 1;
    MoveTo(vX,vY-CrosshairSize);LineTo(vX,vY+CrosshairSize);
    MoveTo(vX-CrosshairSize,vY);LineTo(vX+CrosshairSize,vY);
  end;
  Preview.Color:=Integer(RV)
             OR (Integer(GV) SHL 8)
             OR (Integer(BV) SHL 16);
  ctF.Position:=F; cnF.Text:=IntToStr(F);
  ctH.Position:=H; cnH.Text:=IntToStr(H);
  ctS.Position:=S; cnS.Text:=IntToStr(S);
  ctR.Position:=RV;cnR.Text:=IntToStr(RV);
  ctG.Position:=GV;cnG.Text:=IntToStr(GV);
  ctB.Position:=BV;cnB.Text:=IntToStr(BV);
  sR:=hex[RV SHR 4+1]+hex[RV AND 15+1];
  sG:=hex[GV SHR 4+1]+hex[GV AND 15+1];
  sB:=hex[BV SHR 4+1]+hex[BV AND 15+1];
  csHTML.Text:='#'+sR+sG+sB;
  csDelphi.Text:='$'+sB+sG+sR;
  Lock:=false;
end;

procedure TMainForm.CreateCircle;
var X,Y:integer;
    LX,LY,LR,La:integer;
    CurrLine:PByteArray;
    CurrPos:^byte;
    R,G,B:byte;
begin
  for Y:=0 to 239 do begin
    LY:=Y-120;
    CurrLine:=Circ.ScanLine[Y];
    CurrPos:=@CurrLine[0];
    for X:=0 to 239 do begin
      LX:=X-120;
      LR:=sqr(LX)+sqr(LY);
      if (LR>CircleMinR2) AND (LR<CircleMaxR2) then begin
        if LY<>0 then begin
          La:=round(ArcTan(LX/LY)*rad2deg);
          if LY>0 then inc(La,180);
        end else begin
          if LX>0 then La:=270 else La:=90;
        end;
        if La<0 then inc(La,360);
        if La>359 then dec(La,360);
        R:=SlopeR[La]*255 SHR 15;
        G:=SlopeG[La]*255 SHR 15;
        B:=SlopeB[La]*255 SHR 15;
      end else begin
        R:=192;G:=192;B:=192;
      end;
      CurrPos^:=B;inc(CurrPos);
      CurrPos^:=G;inc(CurrPos);
      CurrPos^:=R;inc(CurrPos);
    end;
  end;
  ColPick.Canvas.Draw(0,0,Circ);
end;

procedure TMainForm.ColPickMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var R:integer;
begin
  if (X>=56) AND (X<184) AND (Y>=56) AND (Y<184) then begin
    HoldMouseOnGradient:=true;
    ColPickMouseMove(Sender,Shift,X,Y);
    exit;
  end;
  R:=sqr(X-120)+sqr(Y-120);
  if (R>CircleMinR2) AND (R<CircleMaxR2) then begin
    HoldMouseOnCircle:=true;
    ColPickMouseMove(Sender,Shift,X,Y);
  end;
end;

procedure TMainForm.ColPickMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  HoldMouseOnCircle:=false;
  HoldMouseOnGradient:=false;
end;

procedure TMainForm.ColPickMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var LX,LY,La:integer;
begin
  if HoldMouseOnCircle then begin
    LX:=X-120;LY:=Y-120;
    if LY<>0 then begin
      La:=round(ArcTan(LX/LY)*rad2deg);
      if LY>0 then inc(La,180);
    end else begin
      if LX>0 then La:=270 else La:=90;
    end;
    if La<0 then inc(La,360);
    if La>359 then dec(La,360);
    F:=La;
    ChangeHue;
  end;
  if HoldMouseOnGradient then begin
    S:=X-56;  S:=(S SHL 1) OR (S SHR 6);
    H:=183-Y; H:=(H SHL 1) OR (H SHR 6);
    if H<0 then H:=0; if H>=256 then H:=255;
    if S<0 then S:=0; if S>=256 then S:=255;
    ChangeColor(FHS2RGB);
  end;
end;

procedure TMainForm.NumbersOnly(Sender: TObject; var Key: Char);
begin
  if not (Key IN [#8,'0'..'9']) then Key:=#0;
end;

procedure TMainForm.cnFChange(Sender: TObject);
begin
  if Lock then exit;
  F:=StrToInt(cnF.Text);
  while F>359 do dec(F,360);
  if F<0 then inc(F,360);
  ChangeHue;
end;

procedure TMainForm.cnHSChange(Sender: TObject);
begin
  if Lock then exit;
  H:=StrToInt(cnH.Text); if H>255 then H:=255;
  S:=StrToInt(cnS.Text); if S>255 then S:=255;
  ChangeColor(FHS2RGB);
end;

procedure TMainForm.ctChange(Sender: TObject);
var p:pointer;
begin
  p:=Pointer(TTrackBar(Sender).Tag);
  TEdit(p^).Text:=IntToStr(TTrackBar(Sender).Position);
end;

procedure TMainForm.ConvRGB2FHS;
var MaxV,MedV,MinV:byte;
    MaxC,MedC:TColorComponent;
    Delta,SubPos:integer;
type TSortItem=array[0..1]of byte;
     TSortTab=array[0..2]of TSortItem;
var SortTab:TSortTab;SortTemp:TSortItem; a,b:byte;
begin
  SortTab[0,0]:=RV;SortTab[0,1]:=byte(ccR);
  SortTab[1,0]:=GV;SortTab[1,1]:=byte(ccG);
  SortTab[2,0]:=BV;SortTab[2,1]:=byte(ccB);
  for a:=0 to 1 do for b:=(a+1) to 2 do
    if SortTab[a,0]<SortTab[b,0] then begin
      SortTemp:=SortTab[a];SortTab[a]:=SortTab[b];SortTab[b]:=SortTemp;
    end;
  MaxV:=SortTab[0,0];MaxC:=TColorComponent(SortTab[0,1]);
  MedV:=SortTab[1,0];MedC:=TColorComponent(SortTab[1,1]);
  MinV:=SortTab[2,0];
  if MaxV=0 then begin
    F:=0;H:=0;S:=0;exit;
  end;
  H:=MaxV;
  Delta:=MaxV-MinV;
  S:=round(255.0*Delta/MaxV);
  if Delta>0 then SubPos:=round((MedV-MinV)/Delta*60.0)
             else SubPos:=0;
  case MedC of
    ccR:if MaxC=ccG then F:=120-SubPos else F:=240+SubPos;
    ccG:if MaxC=ccR then F:=000+SubPos else F:=240-SubPos;
    ccB:if MaxC=ccR then F:=360-SubPos else F:=120+SubPos;
  end;
  if F>360 then dec(F,360);
end;

procedure TMainForm.cnRGBChange(Sender: TObject);
var I:integer;
begin
  if Lock then exit;
  I:=StrToInt(cnR.Text); if I>255 then RV:=255 else RV:=I;
  I:=StrToInt(cnG.Text); if I>255 then GV:=255 else GV:=I;
  I:=StrToInt(cnB.Text); if I>255 then BV:=255 else BV:=I;
  ChangeColor(RGB2FHS);
end;

procedure TMainForm.cpHTMLClick(Sender: TObject);
begin
  Clipboard.AsText:=csHTML.Text;
end;

procedure TMainForm.cpDelphiClick(Sender: TObject);
begin
  Clipboard.AsText:=csDelphi.Text;
end;

procedure TMainForm.HexOnly(Sender: TObject; var Key: Char);
var temp:byte;
begin
  if Key=#13 then begin
    GetHexColor(TEdit(Sender).Text);
    if Sender=csDelphi then begin
      temp:=RV;RV:=BV;BV:=temp;
    end;
    ChangeColor(RGB2FHS);
    exit;
  end;
  Key:=upcase(Key);
  if ((Key>='0') AND (Key<='9')) OR ((Key>='A') AND (Key<='F')) OR
     (Key=char(TEdit(Sender).Tag)) then exit;
  Key:=#0;
end;

procedure TMainForm.GetHexColor(const Raw:str7);
var Pure:string[7]; a,b:byte; c:char;
begin
  Pure:='000000';
  b:=0;
  for a:=1 to length(Raw) do begin
    c:=Raw[a];
    if pos(c,hex)>0 then begin
      inc(b);
      Pure[b]:=c;
    end;
  end;
  RV:=((pos(Pure[1],hex)-1) SHL 4) OR (pos(Pure[2],hex)-1);
  GV:=((pos(Pure[3],hex)-1) SHL 4) OR (pos(Pure[4],hex)-1);
  BV:=((pos(Pure[5],hex)-1) SHL 4) OR (pos(Pure[6],hex)-1);
end;

procedure TMainForm.LoadPalette;
var cnt:integer;
begin
  with PalData[CurrentPal],PalBox.Items do begin
    Clear;
    for cnt:=1 to Entries do
      Add(Data[cnt-1].Name);
  end;
  BAdd.Enabled:=(CurrentPal=0) AND (ColorName.Text<>'');
  BDel.Enabled:=(CurrentPal=0) AND (PalBox.Items.IndexOf(ColorName.Text)>=0);
end;

procedure TMainForm.PalSelectChange(Sender: TObject);
begin
  CurrentPal:=PalSelect.ItemIndex;
  LoadPalette;
end;

procedure TMainForm.PalBoxDrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);
begin
  with PalBox.Canvas,PalData[CurrentPal].Data[Index] do begin
    FillRect(Rect);
    Brush.Color:=Col;
    FillRect(Classes.Rect(Rect.Left,Rect.Top,Rect.Left+16,Rect.Bottom));
    Brush.Style:=bsClear;
    TextOut(Rect.Left+19,Rect.Top+1,Name);
  end;
end;

procedure TMainForm.PalBoxClick(Sender: TObject);
var Item:integer;
begin
  Item:=PalBox.ItemIndex;
  if Item<0 then exit;
  with PalData[CurrentPal].Data[Item] do begin
    RV:=R;
    GV:=G;
    BV:=B;
    ColorName.Text:=Name;
  end;
  ChangeColor(RGB2FHS);
end;

procedure TMainForm.ColorNameChange(Sender: TObject);
begin
  BAdd.Enabled:=(CurrentPal=0) AND (ColorName.Text<>'');
  BDel.Enabled:=(CurrentPal=0) AND (PalBox.Items.IndexOf(ColorName.Text)>=0);
end;

procedure TMainForm.BAddClick(Sender: TObject);
var Index:integer;
begin
  Index:=PalBox.Items.IndexOf(ColorName.Text);
  if Index<0 then PalData[0].Add(RV,GV,BV,ColorName.Text)
             else PalData[0].Replace(Index,RV,GV,BV,ColorName.Text);
  LoadPalette;
end;

procedure TMainForm.BDelClick(Sender: TObject);
var Index:integer;
begin
  Index:=PalBox.Items.IndexOf(ColorName.Text);
  if Index<0 then exit;
  PalData[0].Remove(Index);
  LoadPalette;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  PalData[0].SaveToFile(PaletteFileName);
end;

procedure TMainForm.BScreenClick(Sender: TObject);
begin
  with ScreenShotForm do
    if Visible then Hide else Show;
end;

procedure TMainForm.ImportColor(Color:TColor);
var NewR,NewG,NewB:integer;
begin
  NewR:=Color AND $FF;
  NewG:=(Color SHR 8) AND $FF;
  NewB:=(Color SHR 16) AND $FF;
  if (NewR<>RV) OR (NewG<>GV) OR (NewB<>BV) then begin
    RV:=NewR; GV:=NewG; BV:=NewB;
    ChangeColor(RGB2FHS);
  end;
end;

end.
